import numpy as np
import random

class Individual:
    """
    Individual is a representation of a particular config pin and its 
    voltage commands from a given main start time to an end time.
    During this time there may be several changes to the amplitude
    """
    def __init__(self, total_pins, max_leds, pins, start, end):
        """
        'pins' denotes the pins' numbers
        """
        assert start < int(float(end)/3.0), 'the end time must currently be at least 3 times as large as the start time'
        # start < end is a must, the rest is a hack due to 
        # the same populate function hack
        
        #possibly add an ID

        self.global_start = start
        self.global_end = end
        self.max_leds = max_leds 
        self.pins = pins

        s = self.global_start
        e = self.global_end
        """input_pins and config_pins are only numbers. 
        The actual pins must also be created. 
        """
        self.input_pins = self.random_assign_inputs(pins)
        self.actual_input_pins = {}
        for p in self.input_pins:
            self.actual_input_pins[p] = Pin(p, s, e, "INPUT")
        
        self.config_pins = list(set(pins).difference(set(self.input_pins)))
        self.actual_config_pins = {}
        for p in self.config_pins:
            self.actual_config_pins[p] = Pin(p, s, e, "CONFIG")

        self.available_pins = list( set(total_pins).difference(set(pins)) )
        
    def random_assign_inputs(self, pins):
        """Choose randomly one of the pins to be the input pins """
        index = int(np.random.randint(0, len(pins), 1))
        return  [pins[index]]

    def mutate(self):
        self.mutate_wires()# will maybe change to conditionally 'n' wires mutated

    def mutate_wires(self):

        """no_pins_for_mut generates the number of config pins to be chosen mutation"""
        no_pins_for_mut = int(np.random.randint(1, len(self.config_pins), 1))
        indices_for_mut = random.sample(range(len(self.config_pins)), no_pins_for_mut)
        pins_for_mut = [self.config_pins[i] for i in indices_for_mut]
        #print "Number of pins for mutation: {}\npin indices for mutation{}\npins for mutation{}".format(no_pins_for_mut, indices_for_mut, pins_for_mut)
        for pin in pins_for_mut:
            #print "Mutating pin no: {}".format(idx)
            self.actual_config_pins[pin].mutate_pin()
        #for e in self.actual_config_pins:
        #    print e


    def __str__(self):
        retval = ""
        retval += "Individual\n" + "start: " + str(self.global_start) + "\n" + "end: "
        retval += str(self.global_end) + "\n" + "pins:" + str(self.pins)
        retval += "\n" + "input pin: " + str(self.input_pins) + "\n\n" 
        retval += str(self.actual_input_pins[self.input_pins[0]]) + "\n" 
        retval += "config pins " + str(self.config_pins) + "\n"

        for key in self.actual_config_pins.keys():
            retval += str(self.actual_config_pins[key]) +"\n" 

        return retval 
class Pin:

    def __init__(self, pin,  start, end, pin_type):
        self.global_start = start
        self.global_end = end 
        self.pin_num = pin
        self.cmd_list = []
        #kind of dumb. Consider refactoring later. But to be clear, populate puts commands in cmd_list
        self.populate(self.global_start, self.global_end, pin_type)

    def __str__(self):
        return "Pin: " + str(self.pin_num) + "\n" + "start: " + str(self.global_start) + "\n" + "end: " + str(self.global_end) + "\n" + "command list:\n" + self.stringify_cmd_list()
    
    def stringify_cmd_list(self):
        if not self.cmd_list:
            return ""
        else:
            return str(self.cmd_list[0])

    def mutate_pin(self):
        odds_of_mut = 0.4 # to ease adjustments in the future
        volt_mut_odds = float(np.random.random())
        if volt_mut_odds <= odds_of_mut:
            new_volt = int(np.random.randint(1, 2500, 1))
            while not self.check_voltage(new_volt):
                new_volt = int(np.random.randint(1, 2500, 1))
        
            start = self.cmd_list[0][0]
            end = self.cmd_list[0][1]
            self.cmd_list[0] = (start, end, new_volt)
            
    def populate(self, start, end, pin_type):
        if pin_type == "CONFIG":
            self.populate_config_pins(start, end)

    def populate_config_pins(self, start, end):
        new_volt = int(np.random.randint(1, 2500, 1))
        while not self.check_voltage(new_volt):
            new_volt = int(np.random.randint(1, 2500, 1))
    
        self.cmd_list = [(start, end, new_volt)]
        
    def check_voltage(self, new_volt):
        """the min and max voltages have been found experimentally to be reasonable"""
        max_voltage = 2500
        min_voltage = 1
        if (new_volt < min_voltage or new_volt > max_voltage):
           return False
        else:
            return True
