import sys
from time import *
import numpy
import matplotlib.pyplot as plt
sys.path.append('../../mecobo/Thrift interface/gen-py/NascenseAPI_v01e/')
import emEvolvableMotherboard
from ttypes import *

from bitstring import *
from thrift import Thrift
from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol

#importing own files
from individual import *
import operator
from copy import deepcopy

print "Starting up ea.\nSetting up connections"
#transport = TSocket.TSocket('localhost', 9090)
f = open('server.txt','r')
ip = f.readline().rstrip()
transport = TSocket.TSocket(ip, 9090)
transport = TTransport.TBufferedTransport(transport)

prot = TBinaryProtocol.TBinaryProtocol(transport)
cli = emEvolvableMotherboard.Client(prot);
transport.open();

cli.reset()
cli.clearSequences()

#plt.ylim(-5, 5)
print "Connections complete."


### helper functions
"""
the function 'create_pin_action' is an abstraction over the creation of a pin signal for Mecobo.
Its parameters are as follows:
    pin = pin number, typically 0-15
    start = start time for signal
    end = end time for signal
    freq_or_ampl = tuple ("FREQ" or "AMPL", value) , where value is what the amplitude or frequency is to be set at
    op_type = signal type: "DIGITAL", "CONSTANT" (digital or analog signal is made) or 
    "RECORDING": with the line 'it.waveformType = emWaveFormType.PWM' the recorded signal will be attempted as a digital signal. Else it's analog
    cli = the client stuff: emEvolvableMotherboard.....
    """
def create_pin_action(pin, start, end, freq_or_ampl, op_type, cli):
    it = emSequenceItem()
    it.pin = [pin]
    it.startTime = start
    it.endTime = end
    if freq_or_ampl[0] == "FREQ":
        it.frequency = freq_or_ampl[1]
    elif freq_or_ampl[0] == "AMPL":
        it.amplitude = freq_or_ampl[1]

    if op_type == "DIGITAL":
        it.operationType = emSequenceOperationType().DIGITAL 
    elif op_type == "CONSTANT":
        it.operationType = emSequenceOperationType().CONSTANT
    elif op_type == "RECORDING":
        #it.waveformType = emWaveFormType.PWM
        #PWM will make the recording digital, but output decoding as '1' or '0' is my job
        it.operationType = emSequenceOperationType().RECORD  
    cli.appendSequenceAction(it)

def select_new_generation(fitnesses, inds):
    """
    Might be a bad idea to make this function here.
    Anyway, takes in a list of individuals and a corresponding list of fitness values
    for each individual.
    Current implementation is simple promotion of best individual
    """

    max_idx, maximum = max(enumerate(fitnesses), key=operator.itemgetter(1))

    #max_score = -1.0
    #max_idx = -1
    #for idx,num in enumerate(fitnesses):
    #    if num > max_score:
    #        max_score = num
    #        max_idx = idx
    
    return [deepcopy(inds[max_idx]) for _ in xrange(len(inds))]

def set_rec_and_gnds():
    #solar panel. This is hardcoded and unchangeable
    create_pin_action(rec_sol_panel, start_time, end_time, ("FREQ", sampling_freq) , "RECORDING", cli)
    create_pin_action(gnd_sol_panel, start_time, end_time, ("FREQ",  0) , "DIGITAL", cli)
    #leds gnd
    create_pin_action(gnd_led, start_time, end_time, ("AMPL",  0) , "CONSTANT", cli)

def set_config_pins(conf_pins):
    start_idx = 0
    end_idx = 1
    volt_idx = 2
    offset = 10

    #set up all the pins with their volt parameters
    for pin in conf_pins.keys():
        c_pin = conf_pins[pin]
        for cmd in c_pin.cmd_list:
            c_start = cmd[0] #c_pin.global_start
            c_end = cmd[1]#c_pin.global_end
            c_volt = cmd[2]
            #first set up all pins with the first bit of digital & freq signal
            create_pin_action(c_pin.pin_num, c_start, offset, ("FREQ", 0), "DIGITAL", cli)
            if c_volt == 0:
                create_pin_action(c_pin.pin_num, offset, c_end, ("FREQ", c_volt), "CONSTANT", cli)
            else:
                create_pin_action(c_pin.pin_num, offset, c_end, ("FREQ", c_volt), "DIGITAL", cli)

def set_fitness_test(input_pin, test_input):
    start_idx = 0
    end_idx = 1
    volt_idx = 2

    """Setting low and high frequency respctively to 20Hz and 500Hz """
    treated_inputs = [] 
    for inp in test_input:
        if inp == '0':
            treated_inputs.append(100)
        elif inp == '1':
            treated_inputs.append(1000)
        else:
            print "WTF. Panic"
    
    offset = 10
    no_tests = len(test_input)
    
    key = input_pin.keys()[0]
    pin = input_pin[key]
    create_pin_action(pin.pin_num, pin.global_start, offset, ("FREQ", 0), "DIGITAL", cli)
    first_end = pin.global_end * float(1)/no_tests
    create_pin_action(pin.pin_num, offset, first_end, ("FREQ", treated_inputs[0]), "DIGITAL", cli)

    for i,inp in enumerate(treated_inputs[1:]):
        start = pin.global_end * (float(i+1) / no_tests)
        end = pin.global_end * (float(i+2) / no_tests)
        create_pin_action(pin.pin_num, start, end, ("FREQ", inp), "DIGITAL", cli)


def run_experiment():
    ###Run experiment
    #print "pin actions recorded.\nRunning experiment"
    cli.runSequences()
    cli.joinSequences()

def run_fitness_function(ind):
    """Have set input to "10" for logging. Will do several evo runs and ascertain statistics about the evolutions.
    The specific values that will relate to 0 and 1 respectively will be set in"""
    inputs = "1010"

    output, raw_data = run_fitness_part(ind, inputs)

    #TODO Possibly split the output into sections that are to correspond to the input params.
    #Then evaluate each section of output against that part of inputs and calculate a fitness
    #based on maybe TP*TN /( (TP+FP)(TN+FN) )
    # where T = True, F = False, P = Positive, N = Negative 
    #This puts the fitness values somewhere in [0,1]

    tp,tn,fp,fn = output
    #print "tp:{},tn:{},fp:{},fn:{}".format(tp,tn,fp,fn)
    #fitness = float(tp*tn) / ( (tp+fp) * (tn+fn) )
    #fitness = float(tp*tn) / ( (tp+fp + 1) * (tn+fn + 1) )
    fitness = tp + tn

    #return fitness 
    #first time, maybe just return output to see exactly what I am working with
    return fitness, raw_data, output

def run_fitness_part(ind, fitness_part):

    cli.clearSequences()
    cli.reset()
    
    #print "Setting recording and grounds"
    set_rec_and_gnds()
    #print "Yay they were set"
    set_config_pins(ind.actual_config_pins)
    set_fitness_test(ind.actual_input_pins, fitness_part)
    #print "About to start dah experiment"
    run_experiment()
    res = []
    for i in cli.getRecording(rec_sol_panel).Samples:
        res.append(i * (5.0/4096.0))
    #must probably clear sequences here
    #print "experiment completah. Clean up client and such"
    cli.clearSequences()
    cli.reset()
    """
    At this point choose an output calculation strategy for the result points
    """
    part_fitness = compute_part_output(res, fitness_part)
    return part_fitness, res
 
def compute_part_output(res, fitness_part):
    """Hardcoded approximate ceilings and floors for values here.
    ceiling is unknown 
    floor is unknown

    fitness scoring scheme will likely be 
    1) dividing output into chunks corresponding to tests
    2) Average the output? Score each chunk in relation to how low or high the output is?
       Maybe choose a threshold for low and high, and simply state 1 or 0 from that.
       Then let the total fraction score take over.
    """
    #avg = float(sum(res)) / float(len(res))
    #logical_high_max = 0.267
    #logical_low_floor = -0.015

    #       1)Divide output into chunks that correspond to the tests
    #       Possibly Full_time_of_test / number_of_tests = delta
    #       Then divide output samples by deltas and hope those delta chunks will correspond
    #       2)establish 'high' and 'low' thresholds and do a test towards them:
    #       a)one possibility is to merely average the output and check it for 
    #       proximity to the threshold.
    #       b)look for large portion of contiguous area(s) in correct position(over or under)
    #       and check if correctly positioned area is >= some %. Maybe 60-70. 50 feels
    #       insufficient
    #       c) do threshold crossing rate check. Find some threshold, e.g. 
    
    #"""current scheme is threshold crossing rate. Initial threshold 
    #(i.e. the point of crossing in order to be counted towards the crossing rate) 
    #is set to 0.05 """
    #threshold = 0.05

    ###DEBUG MODE #####

    #print "average value: {}".format(float(sum(res) / float(len(res)) ) )
    #plt.plot(res)
    #plt.show()

    #######

    number_of_tests = len(fitness_part)
    delta = int(float(len(res)) / number_of_tests)

    true_positive = 0
    true_negative = 0
    false_positive = 0
    false_negative = 0

    dominating_frequencies = []
    #spectral_centroids = []

    for test in range(number_of_tests):
        start_idx = delta * test
        end_idx = delta * (test+1)

        """First approx. 330 samples are a weird anomaly in the output. Possibly the solar panel'revving up'"""
        if start_idx == 0:
            start_idx = 330
        current_test = res[start_idx : end_idx+1]
        #taking stare value in the sample and count crossings from there on out

        res_np = np.asarray(current_test)
        spectrum = np.fft.fft(res_np)
        freq = np.fft.fftfreq(len(spectrum)) # calculate the frequencies for the x axis of the plot
        freq = freq * sampling_freq 

        #title_test_text = ""
        #if fitness_part[test] == '1':
        #    title_test_text = "high frequency"
        #else:
        #    title_test_text = "low frequency"

        #plt.title("Fast fourier transform: " + title_test_text)
        #plt.xlabel('Frequencies')
        #plt.ylabel('Magnitude')
        
        #plt.plot(freq[:len(freq)/2+1], abs(spectrum))
        spectrum[0] = 0
        level = max(abs(spectrum))
        mask = (abs(spectrum) >= level)
        peaks = freq[mask]

        dom_freq = abs(peaks[0])
        #print "Dominating frequency and current test {}, {}".format(dom_freq, fitness_part[test])
        #plt.show()

        dominating_frequencies.append(dom_freq)
        
        logic_high = 1000
        logic_low = 100
        assessed_logic_value = -1

        #### Spectral centroid attempt ####
        #magnitudes = np.abs(np.fft.rfft(res_np))
        #spec_centroid = spectral_centroid(freq[:len(freq)/2 + 1], magnitudes)
        #print "Spectral centroid: {}". format(spec_centroid)
        #spectral_centroids.append(spec_centroid)
        
        if dom_freq < ((logic_high  + logic_low) / 2.0):
            assessed_logic_value = 0
        else:
            assessed_logic_value = 1

        

        #fft, = plt.plot(fft_estimate, label="fft'ed slice")
        #raw, = plt.plot(current_test, label="raw output")
        #plt.ylabel('voltage, or ...fft values?')
        #plt.xlabel('samples')
        #plt.title('raw output sample slices and their fft')
        #plt.grid(True)
        #plt.legend([fft, raw])

        #plt.show()

        expected_result = fitness_part[test]
        # have to decide on some crossing rate threshold
        """current crossing rate threshold for 'high' and 'low' is set to 10% on a guess"""
        #cross_rate_thresh = 0.1
        """
        expected | result | output
            1       1       true positive
            1       0       false negative 
            0       1       false positive
            0       0       true negative
        """


        if expected_result == '1':
            if assessed_logic_value:
                true_positive += 1
            else:
                false_negative += 1
        elif expected_result == '0':
            if assessed_logic_value:
                false_positive += 1
            else:
                true_negative += 1
    
    #plt.plot(spectral_centroids)
    #plt.title("Spectral centroids for the subtests of the Fourier transform")
    #plt.xlabel('Test number')
    #plt.ylabel('Magnitude')
    #plt.show()
    
    return true_positive, true_negative, false_positive, false_negative

#def spectral_centroid(freqs, mags):
#    return np.sum(mags*freqs) / np.sum(mags)

def set_darkness(pins):
    for pin in pins:
        create_pin_action(pin, 0, 100000, ("AMPL",  0) , "CONSTANT", cli)

    create_pin_action(rec_sol_panel, 0, 100000, ("FREQ",  10000) , "RECORDING", cli)
    create_pin_action(gnd_sol_panel, 0, 100000, ("FREQ",  0) , "DIGITAL", cli)
    create_pin_action(gnd_led, 0, 100000, ("AMPL",  0) , "CONSTANT", cli)

    run_experiment()
    cli.clearSequences()
    cli.reset()

def check_best_individual(best, iterations):

    fitnesses = []
    
    for i in range(iterations):
        result = run_fitness_function(best)
        set_darkness(best.pins)
        fitnesses.append(result)
    return fitnesses

### intentionally hardcoded values 
start_time = 0
end_time = 2 * 1000000 #1 seconds fitness part runs for now. 
gnd_sol_panel = 15 # wire 15 on the board
rec_sol_panel = 14 # wire 14 on the board
gnd_led = 0
available_leds = range(1,13+1)# the available wires are from 1 to 13
current_max_leds = 8 # due to maximum channel constraints on analog signals 
leds = range(1,current_max_leds) 
individuals = 10 #individuals in the population
max_runs = 30 #maximum allowed number of generations
current_goal = "Tone discriminator"

sampling_freq = 100000 # have to find sampling frequency that mecobo can reliably manage
### ea run
#initialize population: some number of individuals consisting of
# 1 input pin (ouput begin handled by the solar panel)
# 4-5 config pins: each with a list(or something) of amplitudes (or frequencies) and time frames for action during their run

amount_of_evo_runs = 80#60 TODO First, example run to see if things behave as desired. Then 60 runs
print "Running {} evo runs.\nEach at {} population size and {} maximum number of generations.\n".format(amount_of_evo_runs, individuals, max_runs)
for run in range(1, amount_of_evo_runs + 1):

    #print "EA goal: {}".format(current_goal)
    #print "Beginning ea with at most {} generations.".format(max_runs)
    #print "Creating initial individuals with population size {}".format(individuals)
    inds = []
    for i in range(individuals):
        inds.append( Individual( available_leds, current_max_leds, leds, start_time, end_time) )

    print "\n#####\nIndividuals created. Commencing ea number {}".format(run)

    best_fitnesses = []
    worst_fitnesses = []
    #maybe include mediann fitness
    #avg_fitnesses = []
    median_fitnesses = []
    best_max = None
    best_individual = None
    best_individual_raw_output = None
    best_individual_fitness_constituent = None
    main_output_folder = "results/fft/1010_evolutions_MAIN"
    output_folder = "{}/experiment".format(main_output_folder)

    """The evo success variables exist for the sake of logging whether success happened and if so, when """
    evo_success = False
    evo_success_generation = -1

    for generation in range(1, max_runs):
        #evaluate each individual with the fitness function
        fitnesses = []
        raw_output = None
        fitness_value_constituents = []
        
        print "Generation: {}".format(generation)
        for ind in inds:
            fitness, raw_output, fitness_values = run_fitness_function(ind)
            set_darkness(ind.pins)
            fitnesses.append(fitness)
            fitness_value_constituents.append(fitness_values) 

        idx, maximum = max(enumerate(fitnesses), key=operator.itemgetter(1))
        best_fitnesses.append(maximum)
        print maximum
        worst_fitnesses.append(min(fitnesses))
        median_fitnesses.append(np.median(np.array(fitnesses)))

        if best_individual == None or best_max < maximum:
            best_individual = inds[idx]
            best_max = maximum
            best_individual_raw_output = raw_output
            best_individual_fitness_constituents = fitness_value_constituents[idx]
        #else:
        #    if best_max < maximum:
        #        best_individual = inds[idx]
        #        best_max = maximum
        #        best_individual_raw_output = raw_output
        #        best_individual_fitness_constituents = fitness_value_constituents[idx]


        inds = select_new_generation(fitnesses, inds)
        
        for i,ind in enumerate(inds):
            if i == 0:
                continue
            else:
                ind.mutate()
        
        
        max_threshold = 4
        if maximum >= max_threshold:
            evo_success = True
            evo_success_generation = generation
            break

    """Now log the 'winning' individual.
    1)log the config to its own file in a folder for configs
    2)log the fitness value constituents in a single file, each on a new line
    3)log the fitness value in a single file, each on a new line
    4)log evo succcesses and failures in a single file. Counted up
    """
    #config
    with open("{}/configs/run_{}.txt".format(main_output_folder, run), "w") as file:

        output = "{}".format(inds[idx].__str__())
        file.write("{}".format(output))

    #fitness value constituents
    with open("{}/fitness_value_constituents.txt".format(main_output_folder), "ab") as file:

        output = "run {}: {}\n".format(run, best_individual_fitness_constituents)
        file.write("{}".format(output))

    #fitness value
    with open("{}/best_fitnesses.txt".format(main_output_folder), "ab") as file:

        output = "run {}: {}\n".format(run, maximum)
        file.write("{}".format(output))

    #evo successes and failures
    #NOTE evo_successes.txt must exist with 0\n0\n
    """
    It is imperative that the file evo_successes.txt already exists with 0\n0\n already in place
    First number denotes success. Second number denotes failure
    """
    file = open("{}/evo_successes.txt".format(main_output_folder), "r")
    first = file.readline().strip()
    second = file.readline().strip()
    file.close()

    if evo_success:
        first= int(first) + 1
    else:
        second = int(second) + 1
    
    f = open("{}/evo_successes.txt".format(main_output_folder), "w")
    f.close()

    
    file = open("{}/evo_successes.txt".format(main_output_folder), "w")
    output = "{}\n{}\n".format(first, second)
    file.write("{}".format(output))

    #evo success/failure logging with generation stop count for successes
    with open("{}/evo_success_or_failure_log.txt".format(main_output_folder), "ab") as file:
        output = ""
        if evo_success:
            output = "run: {}, evo_success: {}, generation: {}\n".format(run, evo_success, evo_success_generation)
        else:
            output = "run: {}, evo_success: {}\n".format(run, evo_success)

        file.write("{}".format(output))


    print "Ea finished\n"
    cli.clearSequences()
    cli.reset()

print "All Eas finished. Cleaning up and ending connections"
#best, = plt.plot(best_fitnesses, label="best individuals")
cli.clearSequences()
cli.reset()

"""Plot logged runs into graphs and such """

#TODO time to test the best individual to see if the stats hold up

transport.close()

print "Finished"
